import numpy as np
import tkinter as tk
from tkinter import messagebox
from problem.common import get_problem_config
from system.params import *
from system.database.team import TeamDatabase
from system.agent import LoadAgent
from system.utils.network import get_public_ip, get_internal_ip
from system.manager.view import ManagerLoginView, ManagerInitView, ManagerView
from system.manager.task import CreateTaskController, LoadTaskController, RemoveTaskController
from system.manager.access import ManageUserController 


class ManagerController:

    def __init__(self):
        self.root_login = tk.Tk()
        self.root_login.title(f'{TITLE} - Manager')
        self.root_login.protocol('WM_DELETE_WINDOW', self._quit_login)
        self.root_login.resizable(False, False)
        self.view_login = ManagerLoginView(self.root_login)
        self.bind_command_login()

        self.root_init = None
        self.view_init = None

        self.root = None
        self.view = None
        
        self.database = None
        self.table_name = None
        self.table_checksum = None
        self.problem_info = None

        self.refresh_rate = REFRESH_RATE

        self.active_scientist_user = None
        self.active_scientist_host = None
        self.active_technician_user_list = []
        self.active_technician_host_list = []

    def bind_command_login(self):
        '''
        '''
        self.view_login.widget['login'].configure(command=self.login_database)

    def login_database(self):
        '''
        '''
        try:
            user = self.view_login.widget['user'].get()
            passwd = self.view_login.widget['passwd'].get()
            self.database = TeamDatabase('localhost', user, passwd)
        except Exception as e:
            messagebox.showinfo('Error', e, parent=self.root_login)
            return

        self.after_login()

    def bind_command_init(self):
        '''
        '''
        self.view_init.widget['create_task'].configure(command=self.create_task)
        self.view_init.widget['load_task'].configure(command=self.load_task)
        self.view_init.widget['remove_task'].configure(command=self.remove_task)
        self.view_init.widget['manage_user'].configure(command=self.manage_user)

        if self.database.get_current_user() != 'root':
            self.view_init.widget['create_task'].disable()
            self.view_init.widget['remove_task'].disable()
            self.view_init.widget['manage_user'].disable()

    def refresh_info_init(self):
        '''
        '''
        user = self.database.get_current_user()
        public_ip, internal_ip = get_public_ip(), get_internal_ip()
        self.view_init.widget['user'].config(text='Username: ' + user)
        self.view_init.widget['public_ip'].config(text='Public IP: ' + public_ip)
        self.view_init.widget['internal_ip'].config(text='Internal IP: ' + internal_ip)

    def create_task(self):
        '''
        '''
        CreateTaskController(self)
                
    def load_task(self):
        '''
        '''
        LoadTaskController(self)

    def remove_task(self):
        '''
        '''
        RemoveTaskController(self)

    def manage_user(self):
        '''
        '''
        ManageUserController(self)

    def refresh_info(self):
        '''
        '''
        user = self.database.get_current_user()
        public_ip, internal_ip = get_public_ip(), get_internal_ip()
        self.view.widget['man_user'].config(text='Username: ' + user)
        self.view.widget['man_public_ip'].config(text='Public IP: ' + public_ip)
        self.view.widget['man_internal_ip'].config(text='Internal IP: ' + internal_ip)

    def after_login(self):
        '''
        Transit from login window to initial window
        '''
        self._quit_login()
        
        self.root_init = tk.Tk()
        self.root_init.title(f'{TITLE} - Manager')
        self.root_init.protocol('WM_DELETE_WINDOW', self._quit_init)
        self.root_init.resizable(False, False)
        self.view_init = ManagerInitView(self.root_init)
        self.refresh_info_init()
        self.bind_command_init()

        tk.mainloop()

    def after_init(self, table_name):
        '''
        Transit from initial window to main window
        '''
        self._quit_init()

        self.table_name = table_name
        
        self.root = tk.Tk()
        self.root.title(f'{TITLE} - Manager')
        self.root.protocol('WM_DELETE_WINDOW', self._quit)
        self.view = ManagerView(self.root)
        self.refresh_info()

        self.agent = LoadAgent(self.database, self.table_name)
        self.agent.refresh()

        problem_cfg = self.agent.problem_cfg
        if problem_cfg is not None:
            self.view.widget['problem_info'].set_info(problem_cfg)

        self.root.after(self.refresh_rate, self.refresh)

        tk.mainloop()

    def refresh(self):
        '''
        '''
        self.refresh_scientist()
        self.refresh_technician()
        self.refresh_table()
        self.root.after(self.refresh_rate, self.refresh)

    def _quit_login(self):
        '''
        Quit handling for login window
        '''
        self.root_login.quit()
        self.root_login.destroy()

    def _quit_init(self):
        '''
        Quit handling for initial window
        '''
        self.root_init.quit()
        self.root_init.destroy()

    def _quit(self):
        '''
        Quit handling for main window
        '''
        self.root.quit()
        self.root.destroy()

    def run(self):
        tk.mainloop()

    def refresh_scientist(self):
        '''
        '''
        user_list, host_list = self.database.get_active_user_list(return_host=True, role='Scientist')
        
        # TODO: refuse multiple scientist connections

        if self.active_scientist_user is None:
            if user_list == []:
                # offline to offline
                return
            else:
                # offline to online
                self.active_scientist_user = user_list[0]
                self.active_scientist_host = host_list[0]
        else:
            if self.active_scientist_user in user_list:
                # online to online
                return
            else:
                # online to offline
                self.active_scientist_user = None
                self.active_scientist_host = None

        user_label, host_label = self.active_scientist_user, self.active_scientist_host
        if user_label is None: user_label = 'unknown'
        if host_label is None: host_label = 'unknown'
        self.view.widget['sci_user'].config(text='Username: ' + user_label)
        self.view.widget['sci_host'].config(text='Host: ' + host_label)

    def refresh_technician(self):
        '''
        '''
        user_list, host_list = self.database.get_active_user_list(return_host=True, role='Technician')
        if user_list == self.active_technician_user_list and host_list == self.active_technician_host_list: return

        self.active_technician_user_list = user_list
        self.active_technician_host_list = host_list

        # TODO: optimize efficiency
        for i in self.view.widget['tech_disp'].get_children():
            self.view.widget['tech_disp'].delete(i)
        for user, host in zip(user_list, host_list):
            self.view.widget['tech_disp'].insert('', 'end', text=user, values=(host,))

    def refresh_table(self):
        '''
        '''
        assert self.table_name is not None

        if self.view.widget['db_table'] is None:
            if self.agent.check_table_exist():
                columns = self.agent.get_column_names()
                self.view.init_db_table(columns)

                self.agent.refresh()
                self.view.widget['problem_info'].set_info(self.agent.problem_cfg)
            else:
                return

        checksum = self.database.get_checksum(table=self.table_name)
        if checksum == self.table_checksum or checksum == 0: return
        self.table_checksum = checksum

        data = self.database.load_table(name=self.table_name, column=self.view.get_table_columns())
        self.view.widget['db_table'].load(data)
